#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#  E-mail contact: jonas@lbmethod.org
#  The most recent release of Palabos can be downloaded at 
#  <http://www.lbmethod.org/palabos/>
# 
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
# 
#  The library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrapper to Palabos dynamics objects.
"""

__all__ = ['BGK', 'BounceBack']

import datatype

class Dynamics:
    def interface(self, lattice):
        """ Generate an interface for a given lattice, inferring the
            data type from the relaxation parameter of the dynamics. """
        return datatype.lattice_interface (
                lattice, type(self.omega) )
    def plb_dynamics(self, lattice):
        """ Return the corresponding dynamics class from the SWIG
            Palabos interface. """
        interface = self.interface(lattice)
        return self._plb_dynamics(interface)

    def plb_dynamics_enforce_dt(self, interface):
        return self._plb_dynamics(interface)
        
class BGK (Dynamics):
    def __init__(self, omega):
        self.omega=omega
    def _plb_dynamics(self, interface):
        return interface.library.PlbBGKdynamics (
                   interface.swig_dtype(self.omega) )

class BounceBack (Dynamics):
    def __init__(self, vdensity):
        self.vdensity=vdensity
        self.omega=type(vdensity)()
    def _plb_dynamics(self, interface):
        return interface.library.PlbBounceBack (
                   interface.swig_dtype(self.vdensity) )

