#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#  E-mail contact: jonas@lbmethod.org
#  The most recent release of Palabos can be downloaded at 
#  <http://www.lbmethod.org/palabos/>
# 
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
# 
#  The library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
A few utilities for internal use by the Palabos-Python interface files.
"""

import plb
import ppexcept

def block_range(arg, bound):
    if arg is None:
        convert_range = bound
    elif hasattr(arg, '__iter__'):
        convert_range = arg
    else:
        convert_range = [arg, arg]
    return convert_range
    
def block_range2(arg, bound):
    try: 
        if arg.start is None:
            convert_range = bound
        else:
            convert_range = [min(arg.start,arg.stop), max(arg.start,arg.stop)]
    except:
        convert_range = [arg, arg]
    return convert_range
    

def slice_to_box(bounding_box, *slices):
    box = bounding_box.__class__()
    box.x0, box.x1 = block_range2(slices[0], [bounding_box.x0, bounding_box.x1])
    box.y0, box.y1 = block_range2(slices[1], [bounding_box.y0, bounding_box.y1])
    if hasattr(box, 'z0'):  # 3D case
        box.z0, box.z1 = block_range2(slices[2], [bounding_box.z0, bounding_box.z1])
    intersection = bounding_box.__class__()
    if plb.core.intersect( box, bounding_box, intersection):
        box = intersection
    else:
        raise ppexcept.PyPalOutOfRange('Slice does not intersect with the domain of the object.')
    return box

def intersect(box1, box2):
    intersection = box1.__class__()
    plb.core.intersect (box1, box2, intersection)
    return intersection

def boxToList(box):
    try:
        return [box.x0, box.x1, box.y0, box.y1, box.z0, box.z1]
    except:
        return [box.x0, box.x1, box.y0, box.y1]


def reshape_ndarray(array, domain, lastDim):
    try:
        shape = domain.getNx(), domain.getNy(), domain.getNz(), lastDim
    except:
        shape = domain.getNx(), domain.getNy(), lastDim
    array.shape = shape


def assert_ndarray_equals(array, domain, lastDim):
    try:
        shape = domain.getNx(), domain.getNy(), domain.getNz(), lastDim
    except:
        shape = domain.getNx(), domain.getNy(), lastDim
    if array.shape != shape:
        raise ppexcept.PyPalOutOfRange('Array dimensions do not match current subdomain.')

