/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * Functionals for domain initialization -- header file.
 */
#ifndef DYNAMICS_PROCESSOR_3D_H
#define DYNAMICS_PROCESSOR_3D_H

#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "core/dynamics.h"

namespace plb {

/* *************** Class ExternalRhoJcollideAndStream3D ******************* */

template<typename T, template<typename U> class Descriptor>
class ExternalRhoJcollideAndStream3D : public BoxProcessingFunctional3D
{
public:
    // Block 0: lattice; Block 1: rhoBar; Block 2: j.
    virtual void processGenericBlocks( Box3D domain,
                                       std::vector<AtomicBlock3D*> atomicBlocks );
    virtual ExternalRhoJcollideAndStream3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    void collide (
            BlockLattice3D<T,Descriptor>& lattice, Box3D const& domain,
            ScalarField3D<T> const& rhoBarField, Dot3D const& offset1,
            TensorField3D<T,3> const& jField, Dot3D const& offset2, BlockStatistics& stat );
    void bulkCollideAndStream (
            BlockLattice3D<T,Descriptor>& lattice, Box3D const& domain,
            ScalarField3D<T> const& rhoBarField, Dot3D const& offset1,
            TensorField3D<T,3> const& jField, Dot3D const& offset2, BlockStatistics& stat );
    void boundaryStream (
            BlockLattice3D<T,Descriptor>& lattice,
            Box3D const& bound, Box3D const& domain );
};

template<typename T, template<typename U> class Descriptor>
class OnLinkExternalRhoJcollideAndStream3D : public BoxProcessingFunctional3D
{
public:
    // Block 0: lattice; Block 1: rhoBar; Block 2: j.
    virtual void processGenericBlocks( Box3D domain,
                                       std::vector<AtomicBlock3D*> atomicBlocks );
    virtual OnLinkExternalRhoJcollideAndStream3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    void collide (
            BlockLattice3D<T,Descriptor>& lattice, Box3D const& domain,
            ScalarField3D<T> const& rhoBarField, Dot3D const& offset1,
            TensorField3D<T,3> const& jField, Dot3D const& offset2, BlockStatistics& stat );
    void bulkCollideAndStream (
            BlockLattice3D<T,Descriptor>& lattice, Box3D const& domain,
            ScalarField3D<T> const& rhoBarField, Dot3D const& offset1,
            TensorField3D<T,3> const& jField, Dot3D const& offset2, BlockStatistics& stat );
    void boundaryStream (
            BlockLattice3D<T,Descriptor>& lattice,
            Box3D const& bound, Box3D const& domain );
};


}  // namespace plb

#endif  // DYNAMICS_PROCESSOR_3D_H
