/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * A collection of dynamics classes (e.g. BGK) with which a Cell object
 * can be instantiated -- header file.
 */
#ifndef NLD_BOUNDARIES_3D_H
#define NLD_BOUNDARIES_3D_H

#include "core/globalDefs.h"
#include "core/nonLocalDynamics3D.h"
#include "boundaryCondition/boundaryCondition.h"

namespace plb {

template<typename T, template<typename U> class Descriptor>
class ExecuteNonLocalDynamics3D : public BoxProcessingFunctional3D_L<T,Descriptor>
{
public:
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& lattice);
    virtual ExecuteNonLocalDynamics3D<T,Descriptor>* clone() const;
    virtual int getStaticId() const { return staticId; }
private:
    static const int staticId;
};

template<typename T, template<typename U> class Descriptor>
class ExecutePlaneNLD_3D : public BoxProcessingFunctional3D_L<T,Descriptor>
{
public:
    ExecutePlaneNLD_3D();
    ExecutePlaneNLD_3D(int direction_, int orientation_);
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& lattice);
    virtual ExecutePlaneNLD_3D<T,Descriptor>* clone() const;
    virtual int getStaticId() const { return staticId; }
    virtual void serialize(std::string& data) const;
    virtual void unserialize(std::string& data);
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::staticVariables;
    }
private:
    int direction, orientation;
    static const int staticId;
};

template<typename T, template<typename U> class Descriptor>
class ExecuteEdgeNLD_3D : public BoxProcessingFunctional3D_L<T,Descriptor>
{
public:
    ExecuteEdgeNLD_3D();
    ExecuteEdgeNLD_3D(int plane_, int normal1_, int normal2_);
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& lattice);
    virtual ExecuteEdgeNLD_3D<T,Descriptor>* clone() const;
    virtual int getStaticId() const { return staticId; }
    virtual void serialize(std::string& data) const;
    virtual void unserialize(std::string& data);
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::staticVariables;
    }
private:
    int plane, normal1, normal2;
    static const int staticId;
};

template<typename T, template<typename U> class Descriptor>
class ExecuteCornerNLD_3D : public BoxProcessingFunctional3D_L<T,Descriptor>
{
public:
    ExecuteCornerNLD_3D();
    ExecuteCornerNLD_3D(int xNormal_, int yNormal_, int zNormal_);
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& lattice);
    virtual ExecuteCornerNLD_3D<T,Descriptor>* clone() const;
    virtual int getStaticId() const { return staticId; }
    virtual void serialize(std::string& data) const;
    virtual void unserialize(std::string& data);
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::staticVariables;
    }
private:
    int xNormal, yNormal, zNormal;
    static const int staticId;
};


template<typename T, template<typename U> class Descriptor>
void instantiateOuterNLDboundary(MultiBlockLattice3D<T,Descriptor>& lattice, Box3D bbox);

template<typename T, template<typename U> class Descriptor>
void setOuterNLDboundaryDynamics (
        MultiBlockLattice3D<T,Descriptor>& lattice, Dynamics<T,Descriptor>* baseDynamics,
        Box3D bbox, Box3D domain, boundary::BcType bcType );

template<typename T, template<typename U> class Descriptor>
void setOuterNLDboundaryDynamics (
        MultiBlockLattice3D<T,Descriptor>& lattice,
        Dynamics<T,Descriptor>* baseDynamics, Box3D domain, boundary::BcType bcType );

}  // namespace plb

#endif  // NLD_BOUNDARIES_3D_H
