/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef INTERPOLATIONS_2D_HH
#define INTERPOLATIONS_2D_HH

#include "core/globalDefs.h"
#include "core/util.h"
#include "finiteDifference/interpolations2D.h"
#include <vector>

namespace plb {

/* ******** Function linearInterpolationCoefficients ********************* */

template<typename T>
void linearInterpolationCoefficients (
        AtomicBlock2D const& block, Array<T,2> const& position,
        std::vector<Dot2D>& cellPos, std::vector<T>& weights )
{
    cellPos.resize(4);
    cellPos[0] = Dot2D( (plint) position[0],         (plint) position[1]        );
    cellPos[1] = Dot2D( (plint) position[0],         (plint)(position[1]+(T)1.0));
    cellPos[2] = Dot2D( (plint)(position[0]+(T)1.0), (plint) position[1]        );
    cellPos[3] = Dot2D( (plint)(position[0]+(T)1.0), (plint)(position[1]+(T)1.0));

    T u = position[0] - (T)cellPos[0].x;
    T v = position[1] - (T)cellPos[0].y;

    weights.resize(4);
    weights[0] = (1.-u) * (1.-v);
    weights[1] = (1.-u) * (   v);
    weights[2] = (   u) * (1.-v);
    weights[3] = (   u) * (   v);

    PLB_ASSERT( util::fpequal(
                    weights[0] + weights[1] + weights[2] + weights[3],
                    (T)1., 1.e-6) );

    // Convert cell position to local coordinates.
    for (plint iPos=0; iPos<4; ++iPos) {
        cellPos[iPos] -= block.getLocation();
    }
}

}  // namespace plb

#endif  // INTERPOLATIONS_2D_HH
