/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef INTERPOLATIONS_3D_HH
#define INTERPOLATIONS_3D_HH

#include "core/globalDefs.h"
#include "core/util.h"
#include "finiteDifference/interpolations3D.h"
#include <vector>

namespace plb {

/* ******** Function linearInterpolationCoefficients ********************* */

template<typename T>
void linearInterpolationCoefficients (
        AtomicBlock3D const& block, Array<T,3> const& position,
        std::vector<Dot3D>& cellPos, std::vector<T>& weights )
{
    cellPos.resize(8);
    cellPos[0] = Dot3D( (plint) position[0],         (plint) position[1],         (plint) position[2] );
    cellPos[1] = Dot3D( (plint) position[0],         (plint) position[1],         (plint)(position[2]+(T)1.0) );
    cellPos[2] = Dot3D( (plint) position[0],         (plint)(position[1]+(T)1.0), (plint) position[2] );
    cellPos[3] = Dot3D( (plint) position[0],         (plint)(position[1]+(T)1.0), (plint)(position[2]+(T)1.0) );
    cellPos[4] = Dot3D( (plint)(position[0]+(T)1.0), (plint) position[1],         (plint) position[2] );
    cellPos[5] = Dot3D( (plint)(position[0]+(T)1.0), (plint) position[1],         (plint)(position[2]+(T)1.0) );
    cellPos[6] = Dot3D( (plint)(position[0]+(T)1.0), (plint)(position[1]+(T)1.0), (plint) position[2] );
    cellPos[7] = Dot3D( (plint)(position[0]+(T)1.0), (plint)(position[1]+(T)1.0), (plint)(position[2]+(T)1.0) );

    T u = position[0] - (T)cellPos[0].x;
    T v = position[1] - (T)cellPos[0].y;
    T w = position[2] - (T)cellPos[0].z;

    weights.resize(8);
    weights[0] = (1.-u) * (1.-v) * (1.-w);
    weights[1] = (1.-u) * (1.-v) * (   w);
    weights[2] = (1.-u) * (   v) * (1.-w);
    weights[3] = (1.-u) * (   v) * (   w);
    weights[4] = (   u) * (1.-v) * (1.-w);
    weights[5] = (   u) * (1.-v) * (   w);
    weights[6] = (   u) * (   v) * (1.-w);
    weights[7] = (   u) * (   v) * (   w);

    PLB_ASSERT( util::fpequal(
                        weights[0] + weights[1] + weights[2] +
                        weights[3] + weights[4] + weights[5] +
                        weights[6] + weights[7],
                    (T)1., (T)1.e-6) );

    // Convert cell position to local coordinates.
    for (plint iPos=0; iPos<8; ++iPos) {
        cellPos[iPos] -= block.getLocation();
    }
}

}  // namespace plb

#endif  // INTERPOLATIONS_3D_HH
