/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FREE_SURFACE_TENSION_3D_H
#define FREE_SURFACE_TENSION_3D_H

#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "multiBlock/defaultMultiBlockPolicy3D.h"
#include "multiPhysics/freeSurfaceUtil3D.h"
#include "multiPhysics/freeSurfaceInitializer3D.h"
#include "dataProcessors/dataInitializerWrapper3D.h"
#include "offLattice/marchingCube.h"
#include <algorithm>
#include <memory>

namespace plb {

template<typename T,template<typename U> class Descriptor>
class StoreSurfaceNormal3D : public BoxProcessingFunctional3D_ST<T,T,3> {
public:
    virtual void process(Box3D domain, ScalarField3D<T>& volumeFraction, TensorField3D<T,3>& normal);
    virtual StoreSurfaceNormal3D<T,Descriptor>* clone() const {
        return new StoreSurfaceNormal3D(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::nothing;         // Volume fraction.
        modified[1] = modif::staticVariables; // Surface normal.
    }
};

template<typename T>
class IsoFreeSurfaceDefinition3D : public IsoSurfaceDefinition3D<T> {
public:
    virtual bool isInside (
            plint surfaceId, Array<plint,3> const& position ) const;
    virtual bool isValid (
            Array<plint,3> const& position ) const;
    virtual Array<T,3> getSurfacePosition (
            plint surfaceId, Array<plint,3> const& p1, Array<plint,3> const& p2 ) const;
    virtual void setArguments(std::vector<AtomicBlock3D*> const& arguments);
    virtual IsoFreeSurfaceDefinition3D<T>* clone() const;
    // Arg 1: Flags. Arg 2: Volume fraction. Arg 3: Normals.
    virtual plint getNumArgs() const;
    virtual std::vector<plint> getSurfaceIds() const;
private:
    ScalarField3D<int>* flag;
    ScalarField3D<T>* volumeFraction;
    TensorField3D<T,3>* normals;
};


template<typename T>
std::auto_ptr<MultiContainerBlock3D> storeTrianglesOnEdges(MultiScalarField3D<T>& volumeFraction, MultiScalarField3D<int>& flag);


}  // namespace plb

#endif  // FREE_SURFACE_TENSION_3D_H

